@@ -11,7 +11,6 @@
 
 #include <catalog/pg_type.h>
 #include <utils/builtins.h>
-#include <utils/jsonb.h>
 #include <utils/lsyscache.h>
 
 #include <groonga.h>
@@ -20,12 +19,12 @@
 
 #include <string.h>
 
-PGRN_FUNCTION_INFO_V1(pgroonga_match_script_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_match_script_jsonb);
 
 /* v2 */
-PGRN_FUNCTION_INFO_V1(pgroonga_match_jsonb);
-PGRN_FUNCTION_INFO_V1(pgroonga_query_jsonb);
-PGRN_FUNCTION_INFO_V1(pgroonga_script_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_match_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_query_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_script_jsonb);
 
 typedef struct
 {
@@ -70,6 +69,68 @@ static grn_obj *tmpPathsTable = NULL;
 static grn_obj *tmpTypesTable = NULL;
 static grn_obj *tmpValuesTable = NULL;
 
+const char *
+PGrnJSONBIteratorTokenToString(JsonbIteratorToken token)
+{
+	switch (token)
+	{
+	case WJB_DONE:
+		return "done";
+	case WJB_KEY:
+		return "key";
+	case WJB_VALUE:
+		return "value";
+	case WJB_ELEM:
+		return "element";
+	case WJB_BEGIN_ARRAY:
+		return "begin-array";
+	case WJB_END_ARRAY:
+		return "end-array";
+	case WJB_BEGIN_OBJECT:
+		return "begin-object";
+	case WJB_END_OBJECT:
+		return "end-object";
+	default:
+		return "unknown";
+	}
+}
+
+const char *
+PGrnJSONBValueTypeToString(PGRN_JBV_TYPE type)
+{
+	switch (type)
+	{
+	case jbvNull:
+		return "null";
+	case jbvString:
+		return "string";
+	case jbvNumeric:
+		return "numeric";
+	case jbvBool:
+		return "bool";
+	case jbvArray:
+		return "array";
+	case jbvObject:
+		return "object";
+	case jbvBinary:
+		return "binary";
+#ifdef PGRN_HAVE_JSONB_DATETIME
+	case jbvDatetime:
+		return "datetime";
+#endif
+	default:
+		return "unknown";
+	}
+}
+
+Jsonb *
+PGrnJSONBParse(const char *jsonString)
+{
+	Datum jsonbDatum =
+		DirectFunctionCall1(jsonb_in, CStringGetDatum(jsonString));
+	return DatumGetJsonbP(jsonbDatum);
+}
+
 grn_obj *
 PGrnJSONBLookupValuesTable(Relation index,
 						   unsigned int nthAttribute,
@@ -751,8 +812,8 @@ PGrnJSONBInsertContainer(JsonbIterator **iter, PGrnJSONBInsertData *data)
 		default:
 			ereport(ERROR,
 					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg("pgroonga: jsonb iterator returns invalid token: %d",
-							token)));
+					 errmsg("pgroonga: jsonb iterator returns invalid token: %s",
+							PGrnJSONBIteratorTokenToString(token))));
 			break;
 		}
 	}
@@ -821,8 +882,8 @@ PGrnJSONBInsertContainerForFullTextSearch(JsonbIterator **iter,
 		default:
 			ereport(ERROR,
 					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg("pgroonga: jsonb iterator returns invalid token: %d",
-							token)));
+					 errmsg("pgroonga: jsonb iterator returns invalid token: %s",
+							PGrnJSONBIteratorTokenToString(token))));
 			break;
 		}
 	}
@@ -895,13 +956,14 @@ PGrnJSONBCreateFullTextSearchIndexColumn(PGrnCreateData *data,
 	grn_obj *type;
 	grn_obj *lexicon;
 	grn_obj *tokenizer = NULL;
-	grn_obj *normalizer = NULL;
+	grn_obj *normalizers = NULL;
 	grn_obj *tokenFilters = NULL;
 
 	PGrnApplyOptionValues(data->index,
+						  data->i,
 						  PGRN_OPTION_USE_CASE_FULL_TEXT_SEARCH,
 						  &tokenizer, PGRN_DEFAULT_TOKENIZER,
-						  &normalizer, PGRN_DEFAULT_NORMALIZER,
+						  &normalizers, PGRN_DEFAULT_NORMALIZERS,
 						  &tokenFilters,
 						  &flags);
 
@@ -917,7 +979,7 @@ PGrnJSONBCreateFullTextSearchIndexColumn(PGrnCreateData *data,
 							  flags,
 							  type,
 							  tokenizer,
-							  normalizer,
+							  normalizers,
 							  tokenFilters);
 	GRN_PTR_PUT(ctx, data->lexicons, lexicon);
 
@@ -1475,10 +1537,6 @@ PGrnJSONBInsert(Relation index,
 	uint32_t recordSize = 0; /* always 0 */
 	PGrnJSONBInsertData data;
 	unsigned int nthAttribute = 0;
-	Jsonb *jsonb;
-
-	if (isnull[nthAttribute])
-		return recordSize;
 
 	data.index = index;
 	data.isForFullTextSearchOnly =
@@ -1495,8 +1553,11 @@ PGrnJSONBInsert(Relation index,
 	}
 	PGrnJSONBInsertDataInit(&data);
 
-	jsonb = DatumGetJsonbP(values[nthAttribute]);
-	PGrnJSONBInsertJSONB(&data, jsonb);
+	if (!isnull[nthAttribute])
+	{
+		Jsonb *jsonb = DatumGetJsonbP(values[nthAttribute]);
+		PGrnJSONBInsertJSONB(&data, jsonb);
+	}
 
 	PGrnJSONBInsertRecord(index,
 						  sourcesTable,
@@ -1741,8 +1802,8 @@ PGrnSearchBuildConditionJSONContain(PGrnSearchData *data,
 		default:
 			ereport(ERROR,
 					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg("pgroonga: jsonb iterator returns invalid token: %d",
-							token)));
+					 errmsg("pgroonga: jsonb iterator returns invalid token: %s",
+							PGrnJSONBIteratorTokenToString(token))));
 			break;
 		}
 	}
@@ -2019,33 +2080,33 @@ PGrnJSONBRemoveUnusedTables(Oid relationFileNodeID)
 			return;
 	}
 
-		PGrnRemoveJSONValueLexicon("FullTextSearch", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("String", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("Number", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("Boolean", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("Size", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("FullTextSearch", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("String", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("Number", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("Boolean", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("Size", relationFileNodeID);
 
-		{
-			char name[GRN_TABLE_MAX_KEY_SIZE];
+	{
+		char name[GRN_TABLE_MAX_KEY_SIZE];
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONPathsTableNameFormat ".%s",
-					 relationFileNodeID, 0, PGrnIndexColumnName);
-			PGrnRemoveObject(name);
+		snprintf(name, sizeof(name),
+				 PGrnJSONPathsTableNameFormat ".%s",
+				 relationFileNodeID, 0, PGrnIndexColumnName);
+		PGrnRemoveObject(name);
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONValuesTableNameFormat,
-					 relationFileNodeID, 0);
-			PGrnRemoveObject(name);
+		snprintf(name, sizeof(name),
+				 PGrnJSONValuesTableNameFormat,
+				 relationFileNodeID, 0);
+		PGrnRemoveObject(name);
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONPathsTableNameFormat,
-					 relationFileNodeID, 0);
-			PGrnRemoveObject(name);
+		snprintf(name, sizeof(name),
+				 PGrnJSONPathsTableNameFormat,
+				 relationFileNodeID, 0);
+		PGrnRemoveObject(name);
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONTypesTableNameFormat,
-					 relationFileNodeID, 0);
-			PGrnRemoveObject(name);
-		}
+		snprintf(name, sizeof(name),
+				 PGrnJSONTypesTableNameFormat,
+				 relationFileNodeID, 0);
+		PGrnRemoveObject(name);
+	}
 }
