@@ -47,10 +47,10 @@ PGrnWALDisable(void)
 #	include <msgpack.h>
 #endif
 
-PGRN_FUNCTION_INFO_V1(pgroonga_wal_apply_index);
-PGRN_FUNCTION_INFO_V1(pgroonga_wal_apply_all);
-PGRN_FUNCTION_INFO_V1(pgroonga_wal_truncate_index);
-PGRN_FUNCTION_INFO_V1(pgroonga_wal_truncate_all);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_apply_index);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_apply_all);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_truncate_index);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_truncate_all);
 
 #ifdef PGRN_SUPPORT_WAL
 static grn_ctx *ctx = &PGrnContext;
@@ -804,7 +804,7 @@ PGrnWALCreateTable(Relation index,
 				   grn_table_flags flags,
 				   grn_obj *type,
 				   grn_obj *tokenizer,
-				   grn_obj *normalizer,
+				   grn_obj *normalizers,
 				   grn_obj *tokenFilters)
 {
 #ifdef PGRN_SUPPORT_WAL
@@ -838,8 +838,8 @@ PGrnWALCreateTable(Relation index,
 	msgpack_pack_cstr(packer, "tokenizer");
 	msgpack_pack_grn_obj(packer, tokenizer);
 
-	msgpack_pack_cstr(packer, "normalizer");
-	msgpack_pack_grn_obj(packer, normalizer);
+	msgpack_pack_cstr(packer, "normalizers");
+	msgpack_pack_grn_obj(packer, normalizers);
 
 	msgpack_pack_cstr(packer, "token_filters");
 	msgpack_pack_grn_obj(packer, tokenFilters);
@@ -1589,7 +1589,7 @@ PGrnWALApplyCreateTable(PGrnWALApplyData *data,
 	grn_table_flags flags = 0;
 	grn_obj *type = NULL;
 	grn_obj *tokenizer = NULL;
-	grn_obj *normalizer = NULL;
+	grn_obj *normalizers = NULL;
 	grn_obj *tokenFilters = NULL;
 	uint32_t i;
 
@@ -1616,11 +1616,12 @@ PGrnWALApplyCreateTable(PGrnWALApplyData *data,
 														kv,
 														&(buffers->tokenizer));
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "normalizer"))
+		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "normalizer") ||
+				 PGrnWALApplyKeyEqual(context, &(kv->key), "normalizers"))
 		{
-			normalizer = PGrnWALApplyValueGetTableModule(context,
-														 kv,
-														 &(buffers->normalizer));
+			normalizers = PGrnWALApplyValueGetTableModule(context,
+														  kv,
+														  &(buffers->normalizers));
 		}
 		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "token_filters"))
 		{
@@ -1636,7 +1637,7 @@ PGrnWALApplyCreateTable(PGrnWALApplyData *data,
 							flags,
 							type,
 							tokenizer,
-							normalizer,
+							normalizers,
 							tokenFilters);
 }
 
@@ -2147,7 +2148,8 @@ PGrnWALTruncate(Relation index)
 	int64_t nTruncatedBlocks = 0;
 	BlockNumber i;
 	BlockNumber nBlocks;
-	Buffer *buffers;
+	Buffer processingBuffers[MAX_GENERIC_XLOG_PAGES];
+	uint32_t nProcessingBuffers = 0;
 	GenericXLogState *state;
 
 	LockRelation(index, PGrnWALLockMode());
@@ -2158,8 +2160,6 @@ PGrnWALTruncate(Relation index)
 		return 0;
 	}
 
-	buffers = palloc(sizeof(Buffer) * nBlocks);
-
 	state = GenericXLogStart(index);
 
 	{
@@ -2170,7 +2170,7 @@ PGrnWALTruncate(Relation index)
 		buffer = PGrnWALReadLockedBuffer(index,
 										 PGRN_WAL_META_PAGE_BLOCK_NUMBER,
 										 BUFFER_LOCK_EXCLUSIVE);
-		buffers[0] = buffer;
+		processingBuffers[nProcessingBuffers++] = buffer;
 		page = GenericXLogRegisterBuffer(state, buffer, GENERIC_XLOG_FULL_IMAGE);
 		pageSpecial = (PGrnWALMetaPageSpecial *) PageGetSpecialPointer(page);
 		pageSpecial->next = PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1;
@@ -2183,26 +2183,35 @@ PGrnWALTruncate(Relation index)
 		Buffer buffer;
 		Page page;
 
-		buffer = PGrnWALReadLockedBuffer(index, i, BUFFER_LOCK_EXCLUSIVE);
-		buffers[i] = buffer;
-		if (nTruncatedBlocks >= MAX_GENERIC_XLOG_PAGES)
+		if (nProcessingBuffers >= MAX_GENERIC_XLOG_PAGES)
 		{
 			GenericXLogFinish(state);
+			{
+				uint32_t j;
+				for (j = 0; j < nProcessingBuffers; j++)
+				{
+					UnlockReleaseBuffer(processingBuffers[j]);
+				}
+				nProcessingBuffers = 0;
+			}
 			state = GenericXLogStart(index);
 		}
+		buffer = PGrnWALReadLockedBuffer(index, i, BUFFER_LOCK_EXCLUSIVE);
+		processingBuffers[nProcessingBuffers++] = buffer;
 		page = GenericXLogRegisterBuffer(state, buffer, GENERIC_XLOG_FULL_IMAGE);
 		PageInit(page, BLCKSZ, 0);
 
 		nTruncatedBlocks++;
 	}
-
 	GenericXLogFinish(state);
 
-	for (i = 0; i < nBlocks; i++)
 	{
-		UnlockReleaseBuffer(buffers[i]);
+		uint32_t j;
+		for (j = 0; j < nProcessingBuffers; j++)
+		{
+			UnlockReleaseBuffer(processingBuffers[j]);
+		}
 	}
-	pfree(buffers);
 
 	PGrnIndexStatusSetWALAppliedPosition(index,
 										 PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1,
